/* Try to solve a polynomial, using home-grown techniques.
 * By: Dirk Mittler
 * Date: October 25, 2019
 *
 * First Windows-patch added: May 16, 2020
 * Further Windows-patches added until: June 2, 2020
 *
 * Text-based User Interface made easier for Windows users:
 *  March 6, 2021.
 *
 *
 * Refinement:
 * Problem:
 * If given an absurdly large constant term, i.e., (-1e36), and a
 * polynomial of the 12th degree to solve, the program may
 * output some obviously false roots, also with an order of
 * magnitude of ~10^36. The suspected reason this happens is, the
 * purported root being raised to the power of (11), which results
 * in the CPU computing infinity, and assigning this to (epsilond),
 * inside 'search_root()'. Once this corrupted result has been
 * computed, it may both pass and fail whatever set of inequality-tests
 * it needs to, to become validated as a root.
 *
 * Proposed Solution:
 * The test can be performed within 'search_root()', whether
 * (epsilond) is in fact infinite, and if so, to return 'false'.
 *
 * Refinement first conceived on August 22, 2022 by Dirk Mittler.
 * 
 * Afterthought, August 23, 2022:
 * In reality, this malfunction was not purely a failure to test,
 * whether to raise a real number to a power produces infinity.
 * It requires as much that my 'apply_poly()' function, when
 * multiplying complex root-candidates, produces a complex number
 * which is infinite. After all, just as it goes in pure Math, in C++,
 * 
 * (infinity < infinity)
 * 
 * Will produce false.
 * 
 */


#ifdef _WIN32

#define WINVER 0x0601
#define _WIN32_WINNT 0x0601

#define M_PI 3.14159265358979323846264338328
/*
 * Obviously, because the number written above is only a
 * 'double' Literal, the full series of 30 decimal places
 * will not get used. The equivalent precision is to
 * 14 decimal places.
 *
 */

#endif


#include <cmath>
#include <complex>
#include <algorithm>
#include <iostream>
#include <iomanip>
#include <string>
#include <stdlib.h>

// Bug fix added May 14, 2022...
#include <limits>

#define MIN_A0 0.01


#ifndef max
#define max(a, b) ((a > b) ? a : b)
#endif

using std::complex;
using namespace std::complex_literals;
using std::endl;
using std::cout;
using std::setprecision;

using std::string;
using std::cin;
using std::stod;
using std::getline;


typedef complex<long double> CC;

struct Polynomial {
    CC *coeffs;
    int deg;
    int offset;
    CC *roots;
    int solved;
    int o_deg;
};

void normalize(Polynomial *a)
{
    int array_end = a->offset + a->deg;
    CC first_coeff = a->coeffs[a->offset];
    a->coeffs[a->offset] = 1.0;
    for (int i = a->offset + 1; i <= array_end; i++) {
        a->coeffs[i] /= first_coeff;
    }

    return;
}

//  Apply polynomial to parameter.
CC apply_poly(Polynomial *a, CC param)
{

    int array_end = a->offset + a->deg;

    CC power = 1.0;
    CC result = (a->coeffs)[array_end];
    for (int i = 1; i <= a->deg; i++) {
        power = 1.0;
        for (int j = 1; j <= i; j++) {
            power *= param;
        }

        result += power * (a->coeffs)[array_end - i];
    }

    return result;
}

//  In-place division of polynomial.
//  Returns degree of remaining plynomial in same array.
int divide(Polynomial *a, CC root, CC *remaind)
{
    CC multiple = 0;
    int array_end = a->offset + a->deg;
    for (int i = a->offset; i < array_end; i++) {
        multiple = a->coeffs[i];
        a->coeffs[i+1] += multiple * root;
    }

    *remaind = a->coeffs[array_end];
    a->deg--;

    return a->deg;
}

//  Augmented division of polynomial,
//  may or may not polish.
bool try_root(Polynomial *a, CC *root, bool polish, \
              long double epsilon = 1.0E-12)
{
    CC adjustment = 0.0;
    CC remaind = 0.0;
    CC deriv = 0.0;

    if (a->deg < 1) return false;
    int N = a->deg;
    double deriv2_a0 = 0.0;

    int passes = 1;
    if (N == 1) polish = false;
    if (polish) passes = 3;
    int array_end = a->offset + N;

    Polynomial quot;
    quot.coeffs = new CC[array_end + 1];

    while (passes-- > 0) {
		
		using namespace std;
		
        *root += adjustment;

        quot.offset = a->offset;
        quot.deg = N;

        for (int i = 0; i <= array_end; i++) {
            quot.coeffs[i] = a->coeffs[i];
        }

        //  (1.0)^0 is possible here.
        //	!! Some libraries' power functions cast
        //	their first arguments to 'double' (from 'long double').
        deriv2_a0 = N * N * pow(norm(*root) + 1.0, N - 1);
        //	If what was computed here is infinite, verification
        //	will not take place, but, polishing still will.
        
        //	This will be treated as an invalid result...
		#ifdef isinf
        if (isinf(deriv2_a0)) {
            delete[] quot.coeffs;
			return false;
		}
		#endif

        if (norm(apply_poly(&quot, *root)) > \
                epsilon * epsilon * deriv2_a0) {
            delete[] quot.coeffs;
            return false;
        }

        divide(&quot, *root, &remaind);

        adjustment = CC(0.0, 0.0);

        deriv = -apply_poly(&quot, *root);

        if (polish && norm(deriv) > abs(remaind)) {
            adjustment = remaind / deriv;
        }

    }

    for (int i = 0; i < array_end; i++) {
        a->coeffs[i] = quot.coeffs[i];
    }
    a->deg = quot.deg;
    a->offset = quot.offset;

    delete[] quot.coeffs;

    return true;
}

//  Solver for linear case.
//  Assumes that the equation has been normalized.
inline bool solve_lin(Polynomial *a, CC *root)
{
    if (a->deg != 1) return false;

    *root = - a->coeffs[ a->offset + 1 ] ;
    return true;
}

//  Do the actual work.
bool search_root(Polynomial *a, CC *root, long double epsilon = 1.0E-12)
{

    int N = a->deg;

    if (N < 1) return false;

    //  !!
    if (N == 1) return solve_lin(a, root);

    CC xa = -1.0;
//	CC xb = exp(CC(0.0, M_PI / (N + 1)));
    CC xb = exp((M_PI / (N + 1)) * 1i);

    CC ya = apply_poly(a, xa);
    CC yb = apply_poly(a, xb);

    double spread = abs(yb - ya);
    double spread2 = 0.0;

    if (N <= 4 && pow(spread, N) < epsilon) {
        return false;
    } else if (N > 4 && pow(spread, N - 2) < N * N * epsilon) {
        return false;
    }

    int misses = 0;

    long double epsilond = epsilon;
    CC t = 0.0;
    CC xt = 0.0;

    //  Really bad form: while (true)
    while (true) {
		
		using namespace std;
		
        epsilond = N * (pow(abs(xa), N - 1)) * epsilon;
        if (epsilond < epsilon / N)
            epsilond = epsilon / N;

		#ifdef isinf
        if (isinf(epsilond))
            return false;
        #endif
        
        /* At this point in program flow, 3 situations could exist...
         * 
         * Power function does not cast 'long double' to 'double':
         * 
         * 'long double' can hold the powers of root candidates
         * without becoming infinite. Arithmetic continues as though
         * Algebra.
         * 
         * sizeof(double) == sizeof(long double):
         * 
         * (epsilond * epsilond) could be infinite, but if it is,
         * norm(ya) would need to be finite to generate
         * root vicinity. In that case, a hypothetical abs(ya)
         * would also be less than (epsilond).
         * 
         * sizeof(double) < sizeof(long double):
         * 
         * Because (epsilond) holds a value which is finite according
         * to 'double', (epsilond * epsilond) will be finite according
         * to 'long double'. No problem here, either.
         * 
         */

        if (norm(ya) < epsilond * epsilond) {
            *root = xa;
            return true;
        }

        if (norm(ya - yb) <= epsilon * epsilon * epsilon) {
            return false;
        }

        t = (-ya) / (yb - ya);

        xt = (t * xb) + ((CC(1.0, 0.0) - t) * xa);

        xb = xt + (CC(0.0, 0.5) * (xa - xt));
        xa = xt;

        ya = apply_poly(a, xa);
        yb = apply_poly(a, xb);

        spread2 = abs(ya);

        if (spread2 >= spread) {
            misses++;
        }

        if ((misses > N + 2) || (spread2 != spread2)) {
            return false;
        }

        spread = spread2;

    }
}

//  Return number of derivatives...
int create_deriv_list(Polynomial a, Polynomial **l)
{

    CC coeff = 0.0;
    *l = new Polynomial[max(a.deg, 1)];

    Polynomial *L = *l;

    int g_offset = a.offset;
    int array_end = a.deg + g_offset;
    L[0].deg = a.deg;
    L[0].o_deg = a.deg;
    L[0].offset = g_offset;
    L[0].coeffs = new CC[array_end + 1];
    L[0].solved = 0;
    L[0].roots = new CC[a.deg];

    for (int i = 0; i <= array_end; i++) {
        L[0].coeffs[i] = a.coeffs[i];
    }

    if (abs(L[0].coeffs[g_offset]) < MIN_A0) {
        L[0].coeffs[g_offset] = CC(1.0, 0.0);
    }

    normalize(L);

    for (int i = 1; i < a.deg; i++) {
        L[i].deg = L[i-1].deg - 1;
        L[i].o_deg = L[i].deg;
        L[i].offset = g_offset;
        L[i].coeffs = new CC[array_end + 1];
        L[i].solved = 0;
        L[i].roots = new CC[L[i].deg];
        for (int j = 0; j < L[i-1].deg; j++) {

            coeff = L[i-1].coeffs[L[i].deg + g_offset - j];
            L[i].coeffs[L[i].deg + g_offset - j] = coeff * CC(1.0 + j, 0.0);

        }

        normalize(L + i);

    }

    return max(a.deg - 1, 0);
}

void destroy_deriv_list(Polynomial *l)
{
    int array_num = l[0].o_deg;
    for (int i = 0; i < array_num; i++) {
        delete[] l[i].coeffs;
        delete[] l[i].roots;
    }

    delete[] l;
    return;
}

bool climb_root(Polynomial *L, CC *root, int i, int mult, \
                long double epsilon = 1.0E-12)
{
    int s = 0;
    int mult_temp = 0;
    bool last_attempt = false;

    if (i < 0) return true;

    while (( last_attempt = try_root(L + i, root, false, epsilon) )) {
        s = L[i].solved;
        L[i].roots[s] = *root;
        L[i].solved++;
        if (++mult_temp >= mult) break;
    }

    if ( ! last_attempt ) return false;

    return climb_root(L, root, i - 1, mult + 1, epsilon);
}

// Define absolute of real number.
// Should not be necessary.
template<class T>
T abs_d(T x)
{
    if (x < 0) return -x;
    return x;
}

//  Finally, the driver:
bool poly_solve(Polynomial *a, Polynomial **PL, bool polish)
{
    create_deriv_list(*a, PL);
    int s = 0;
    long double epsilon = 1.0E-12;
    Polynomial *L = *PL;

    CC current_root = 0.0;
    CC current_conjugate = 0.0;

    if (sizeof(CC) == 32) {
        epsilon = 1.0L;
        int limits_10 = std::numeric_limits<long double>::digits10 ;
        for (int i = 3 ; i < limits_10 ; i++)
            epsilon /= 10.0L;
    }

    for (int i = a->deg - 1; i >= 0; i--) {
        while ( search_root(L + i, &current_root, epsilon) && \
                try_root(L + i, &current_root, polish, epsilon) ) {

            s = L[i].solved;
            L[i].roots[s] = current_root;
            L[i].solved++;

            climb_root(L, &current_root, i - 1, 2, epsilon);

            if (abs_d(imag(current_root)) > epsilon) {

                current_conjugate = conj(current_root);

                if ( try_root(L + i, &current_conjugate, false, epsilon) ) {

                    s = L[i].solved;
                    L[i].roots[s] = current_conjugate;
                    L[i].solved++;

                    climb_root(L, &current_conjugate, i - 1, 2, epsilon);
                }
            }
        }
    }

    //  Improvised feature:
    //  Print solutions.
    //  This wouldn't be done in actual software.
    //  I.e., Edit out.

    char instring[] = ".";

    cout << endl;
    for (int i = 0; i < L[0].solved; i++) {
        current_root = L[0].roots[i];

        //  Secondary Polishing...
        if (abs_d(imag(current_root)) <= epsilon) {
            current_root = CC(real(current_root), 0.0);
        }
        if (abs_d(real(current_root)) <= epsilon) {
            current_root = CC(0.0, imag(current_root));
        }

        cout << real(current_root) << "  +  ";
        cout << imag(current_root) << "I";
        cout << endl;
    }

    cout << endl;
    cout << endl;
    cout << "Type Enter, to terminate this program." << endl;
    std::cin.get(instring, 1);

    return true;
}

int main(int argc, char *argv[])
{

    Polynomial *PL = NULL;

    Polynomial A;
    int coeff_count = 0;
    int input_length = 1;

    string in_buff = "2";
    std::size_t offset = 0;
    bool polish = true;

    cout << setprecision(14);

    if (sizeof(CC) == 32) {
        cout << setprecision(std::numeric_limits<long double>::digits10 - 2);
    }

//  The following line may be edited out:
    cout << "Roots_Dirk_Mittler, Copyright (C) 2020, Dirk Mittler" << endl;

//  The following lines fulfill 'Appropriate License Notification':
    cout << "This program comes with ABSOLUTELY NO WARRANTY." << endl;
    cout << "This is free software, and you are welcome to redistribute it" << endl;
    cout << "under certain conditions." << endl;
    cout << "The license under which this program has been copyrighted, can be found" << endl;
    cout << "At:  https://www.gnu.org/licenses/gpl-3.0.txt" << endl;
    cout << endl;

    cout << "Type in the degree of the polynomial, followed by" << endl;
    cout << "one coefficient per line, staring with the highest" << endl;
    cout << "power. Enter an empty line to abort." << endl;

    getline(cin, in_buff);

    A.offset = 0;

    try {
        A.deg = stod(&in_buff[0], &offset);
    } catch(...) {
        return -1;
    }

    coeff_count = A.deg;

    if (A.deg < 1) return -1;

    cout << "Polynomial of degree " << A.deg << "..." << endl;

    A.coeffs = new CC[A.deg + 1];

    while (input_length > 0 && coeff_count >= 0) {
        cout << "Coefficient of x^" << coeff_count << " >";
        getline(cin, in_buff);
        input_length = in_buff.length();
        if (input_length > 0) {

            try {
                A.coeffs[A.deg - coeff_count] = CC(stod(&in_buff[0], &offset), 0.0);
            } catch(...) {
                delete[] A.coeffs;
                return -1;
            }

/*
			if (coeff_count == 0) {
				cout << "Overriding the last coefficient with negative infinity.";
				cout << endl;
                A.coeffs[A.deg - coeff_count] =
					CC(-std::numeric_limits<double>::infinity(), 0.0);
			}
*/

            coeff_count -= 1;
        }
    }

    if (input_length == 0) {
        delete[] A.coeffs;
		return -1;
	}

    poly_solve(&A, &PL, polish);

    destroy_deriv_list(PL);

    delete[] A.coeffs;

    return 0;
}
